/* This file is part of smap.
   Copyright (C) 2006, 2007, 2010 Sergey Poznyakoff

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 3, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>. */

#include "smapd.h"
#include <regex.h>
#include <fnmatch.h>

struct smap_sockaddr {
	unsigned netmask;
	int salen;
	struct sockaddr sa;
};

enum query_cond_type {
	query_cond_not,
	query_cond_source,
	query_cond_server,
	query_cond_map
};

enum map_comparison {
	map_eq,
	map_like,
	map_re
};

struct map_cond {
	enum map_comparison op;
	char *str;
	regex_t re;
};
		
struct query_cond {
	struct query_cond *next;
	enum query_cond_type type;
	union {
		struct query_cond *subcond;       /* query_cond_not */
		struct smap_sockaddr *addr;       /* query_cond_source */
		char *id;                         /* query_cond_server */
		struct map_cond mapcond;          /* query_cond_map */
	} v;
};

struct query_route {
	struct query_route *prev, *next;
	char *file;
	unsigned line;
	struct query_cond *cond;
	char *dbname;
	struct smap_database_instance *dbi;
};


static struct query_cond *
query_cond_new(enum query_cond_type type)
{
	struct query_cond *p = ecalloc(1, sizeof(*p));
	p->type = type;
	return p;
}

static struct smap_sockaddr *
smap_sockaddr_new(int family, int len)
{
	struct smap_sockaddr *p = ecalloc(1, sizeof(*p));
	p->salen = len;
	p->sa.sa_family = family;
	return p;
}


#define S_UN_NAME(sa, salen) \
	((salen < offsetof (struct sockaddr_un,sun_path)) ? "" : (sa)->sun_path)

int
match_sockaddr(struct smap_sockaddr *sptr, struct sockaddr const *sa, int len)
{
	if (sptr->sa.sa_family != sa->sa_family)
		return 0;

	switch (sptr->sa.sa_family) {
	case AF_INET:
	{
		struct sockaddr_in *sin_clt = (struct sockaddr_in *)sa;
		struct sockaddr_in *sin_item = (struct sockaddr_in *)&sptr->sa;

		if (sin_item->sin_addr.s_addr ==
		    (ntohl (sin_clt->sin_addr.s_addr) & sptr->netmask))
			return 1;
		break;
	}

	case AF_UNIX:
	{
		struct sockaddr_un *sun_clt = (struct sockaddr_un *)sa;
		struct sockaddr_un *sun_item = (struct sockaddr_un *)&sptr->sa;

		if (S_UN_NAME (sun_clt, len)[0]
		    && S_UN_NAME (sun_item, sptr->salen)[0]
		    && strcmp (sun_clt->sun_path, sun_item->sun_path) == 0)
			return 1;
	}
	}
	return 0;
}


DCLLIST(query, query_route);

enum query_tok {
	T_FROM,
	T_TO,
	T_MAP,
	T_LIKE,
	T_EQ,
	T_DB,
	T_NOT
};

static struct smap_kwtab query_kwtab[] = {
	{ "from", T_FROM },
	{ "to", T_TO },
	{ "map", T_MAP },
	{ "like", T_LIKE },
	{ "eq", T_EQ },
	{ "is", T_EQ },
	{ "uses", T_DB },
	{ "database", T_DB },
	{ "not", T_NOT },
	{ NULL }
};

static char **input;

static char *
nextarg()
{
	if (!*input) {
		smap_error("%s:%u: unfinished query",
			   cfg_file_name, cfg_line);
		return NULL;
	}
	return *input++;
}


static int
parse_query_to(struct query_cond **pcond)
{
	struct query_cond *cond;
	char *arg = nextarg();
	if (!arg)
		return 1;
	cond = query_cond_new(query_cond_server);
	cond->v.id = estrdup(arg);
	*pcond = cond;
	return 0;
}

static int parse_query_not(struct query_cond **pcond);
static int parse_query_map(struct query_cond **pcond);
static int parse_query_from(struct query_cond **pcond);

static int
parse_subcond(struct query_cond **pcond)
{
	char *s;
	int tok;

	s = nextarg();
	if (!s)
		return 1;
	if (smap_kwtab_nametotok(query_kwtab, s, &tok)) {
		smap_error("%s:%lu: unknown keyword: %s",
			   cfg_file_name, cfg_line, s);
		return 1;
	}
	switch (tok) {
	case T_NOT:
		return parse_query_not(pcond);
			
	case T_FROM:
		return parse_query_from(pcond);
			
	case T_TO:
		return parse_query_to(pcond);
			
	case T_MAP:
		return parse_query_map(pcond);
	}
	smap_error("%s:%lu: unexpected keyword: %s",
		   cfg_file_name, cfg_line, s);
	smap_error("%s:%lu: expected one of: not, from, to, map",
		   cfg_file_name, cfg_line);
	return 1;
};
	
static int
parse_query_not(struct query_cond **pcond)
{
	struct query_cond *cond, *subcond;
	if (parse_subcond(&subcond))
		return 1;
	cond = query_cond_new(query_cond_not);
	cond->v.subcond = subcond;
	*pcond = cond;
	return 0;
}

static int
parse_query_map(struct query_cond **pcond)
{
	struct query_cond *cond;
	struct map_cond mapcond;
	char *s = nextarg();
	int tok;
	
	if (!s)
		return 1;

	if (smap_kwtab_nametotok(query_kwtab, s, &tok) == 0) {
		switch (tok) {
		case T_LIKE:
			s = nextarg();
			if (!s)
				return 1;
			mapcond.op = map_like;
			break;

		case T_EQ:
			s = nextarg();
			if (!s)
				return 1;
			mapcond.op = map_eq;
			break;

		default:
			mapcond.op = map_eq;
		}
	} else if (ispunct(s[0])) {
		int rc;
		char *buf;
		size_t len;
		int flags = REG_NOSUB | REG_EXTENDED;
		char *p = strrchr(s, s[0]);
		if (p == s) {
			smap_error("%s:%lu: unfinished regexp: %s",
				   cfg_file_name, cfg_line, s);
			return 1;
		}
		if (p[1]) {
			char *q;
			for (q = p + 1; *q; q++) {
				switch (*q) {
				case 'i':
					flags |= REG_ICASE;
					break;
				case 'b':
					flags &= ~REG_EXTENDED;
					break;
				default:
					smap_error("%s:%lu: uknown regexp "
						   "flag: %c",
						   cfg_file_name, cfg_line,
						   *q);
					return 1;
				}
			}
		}
		len = p - s - 1;
		buf = emalloc(len + 1);
		memcpy(buf, s + 1, len);
		buf[len] = 0;
		rc = regcomp(&mapcond.re, buf, flags);
		free(buf);
		if (rc) {
			char errbuf[512];
			regerror(rc, &mapcond.re, errbuf, sizeof(errbuf));
			smap_error("%s:%lu: regexp error: %s",
				   cfg_file_name, cfg_line, errbuf);
			return 1;
		}
		mapcond.op = map_re;
	} else
		mapcond.op = map_eq;
	mapcond.str = estrdup(s);
	cond = query_cond_new(query_cond_map);
	cond->v.mapcond = mapcond;
	*pcond = cond;
	return 0;
}

static int
parse_query_from(struct query_cond **pcond)
{
	struct query_cond *cond;
	struct smap_sockaddr *sptr;
	char *string = nextarg();
	if (!string)
		return 1;

	if (string[0] == '/') {
		size_t len;
		struct sockaddr_un *s_un;

		len = strlen (string);
		if (len >= sizeof(s_un->sun_path)) {
			smap_error("%s:%u: socket name too long: `%s'",
				   cfg_file_name, cfg_line, string);
			return 1;
		}
		sptr = smap_sockaddr_new(AF_UNIX, sizeof(s_un));
		s_un = (struct sockaddr_un *) &sptr->sa;
		memcpy(s_un->sun_path, string, len);
		s_un->sun_path[len] = 0;
	} else {
		struct in_addr addr;
		struct sockaddr_in *s_in;
		char *p = strchr(string, '/');

		if (p)
			*p = 0;

		if (inet_aton(string, &addr) == 0) {
			struct hostent *hp = gethostbyname(string);
			if (!hp) {
				smap_error("%s:%u: cannot resolve host "
					   "name: `%s'",
					   cfg_file_name, cfg_line, string);
				if (p)
					*p = '/';
				return 1;
			}
			memcpy(&addr.s_addr, hp->h_addr, sizeof(addr.s_addr));
		}
		addr.s_addr = ntohl(addr.s_addr);

		sptr = smap_sockaddr_new(AF_INET, sizeof(s_in));
		s_in = (struct sockaddr_in *) &sptr->sa;
		s_in->sin_addr = addr;

		if (p) {
			*p++ = '/';
			char *q;
			unsigned netlen;

			netlen = strtoul(p, &q, 10);
			if (*q == 0) {
				if (netlen == 0)
					sptr->netmask = 0;
				else {
					sptr->netmask = 0xfffffffful >> (32 - netlen);
					sptr->netmask <<= (32 - netlen);
				}
			} else if (*q == '.') {
				struct in_addr addr;

				if (inet_aton(p, &addr) == 0) {
					smap_error("%s:%u: invalid "
						   "netmask: `%s'",
						   cfg_file_name, cfg_line, p);
					return 1;
				}
				sptr->netmask = addr.s_addr;
			} else {
				smap_error("%s:%u: invalid netmask: `%s'",
					   cfg_file_name, cfg_line, p);
				return 1;
			}
		} else
			sptr->netmask = 0xfffffffful;
	}
	cond = query_cond_new(query_cond_source);
	cond->v.addr = sptr;
	*pcond = cond;
	return 0;
}

int
parse_complex_query()
{
	int rc = 0;
	struct query_cond *head = NULL, *tail = NULL;
	char *dbname = NULL;
	struct query_route *rp;
	
	while (*input && rc == 0) {
		char *s = *input++;
		int tok;
		struct query_cond *cond = NULL;
		
		if (smap_kwtab_nametotok(query_kwtab, s, &tok)) {
			smap_error("%s:%lu: unknown keyword: %s",
				   cfg_file_name, cfg_line, s);
			rc = 1;
			break;
		}
		
		switch (tok) {
		case T_NOT:
			rc = parse_query_not(&cond);
			break;
			
		case T_FROM:
			rc = parse_query_from(&cond);
			break;
			
		case T_TO:
			rc = parse_query_to(&cond);
			break;
			
		case T_MAP:
			rc = parse_query_map(&cond);
			break;
			
		case T_DB:
			s = nextarg();
			if (!s)
				rc = 1;
			else if (dbname) {
				smap_error("%s:%lu: database specified twice",
					   cfg_file_name, cfg_line);
				rc = 1;
			} else
				dbname = estrdup(s);
		}
		if (rc == 1)
			break;
		if (cond) {
			if (tail)
				tail->next = cond;
			else
				head = cond;
			tail = cond;
		}
	}

	if (!dbname) {
		smap_error("%s:%lu: database not specified",
			   cfg_file_name, cfg_line);
		/* FIXME: Free collected data */
		return 1;
	}
	rp = ecalloc(1, sizeof(*rp));
	rp->file = estrdup(cfg_file_name);
	rp->line = cfg_line;
	rp->cond = head;
	rp->dbname = dbname;
	query_attach(rp);
	return 0;
}

int
parse_default_query()
{
	struct query_route *rp;
	int tok;
	char *s = nextarg();
	if (!s)
		return 1;
	if (smap_kwtab_nametotok(query_kwtab, s, &tok) || tok != T_DB) {
		smap_error("%s:%lu: expected `database' but found `%s'",
			   cfg_file_name, cfg_line, s);
		return 1;
	}
	s = nextarg();
	if (!s)
		return 1;
	if (*input) {
		smap_error("%s:%lu: garbage after database name",
			   cfg_file_name, cfg_line);
		return 1;
	}
	rp = ecalloc(1, sizeof(*rp));
	rp->file = estrdup(cfg_file_name);
	rp->line = cfg_line;
	rp->cond = NULL;
	rp->dbname = estrdup(s);
	query_attach(rp);
	return 0;
}

int
parse_query(char **wordv)
{
	input = wordv;

	if (*input && strcmp(*input, "default") == 0) {
		input++;
		return parse_default_query();
	}
	return parse_complex_query();
}

void
link_queries()
{
	struct query_route *p;

	for (p = query_head; p; ) {
		struct query_route *next = p->next;
		struct smap_database_instance *dbi;
		
		dbi = database_locate(p->dbname);
		if (!dbi) {
			smap_error("%s:%u: no such database: %s",
				   p->file, p->line, p->dbname);
			debug(DBG_MODULE, 1, ("removing query %s:%u",
					      p->file, p->line));
			query_detach(p);
			/* FIXME: free memory */
		}
		p->dbi = dbi;
		p = next;
	}
}

struct query_pack {
	const char *server_id;
	struct smap_conninfo const *conninfo;
	const char *map;
	const char *key;
};

static int
match_map(struct map_cond *cond, const char *map)
{
	switch (cond->op) {
	case map_eq:
		return strcmp(cond->str, map) == 0;

	case map_like:
		return fnmatch(cond->str, map, 0) == 0;

	case map_re:
		return regexec(&cond->re, map, 0, NULL, 0) == 0;
	}
	return 0;
}

static int
match_cond(struct query_cond *cond, struct query_pack *qp)
{
	switch (cond->type) {
	case query_cond_not:
		return match_cond(cond->v.subcond, qp);

	case query_cond_source:
		return match_sockaddr(cond->v.addr, qp->conninfo->src,
				      qp->conninfo->srclen);

	case query_cond_server:
		/* FIXME: Implement like/regex */
		return strcmp(cond->v.id, qp->server_id) == 0;

	case query_cond_map:
		return match_map(&cond->v.mapcond, qp->map);
	}
	return 0;
}

static int
match_cond_list(struct query_cond *cond, struct query_pack *qp)
{
	for (; cond; cond = cond->next)
		if (!match_cond(cond, qp))
			return 0;
	return 1;
}

static struct query_route *
find_query_route(struct query_pack *qp)
{
	struct query_route *p;

	for (p = query_head; p; p = p->next) {
		debug(DBG_QUERY, 2, ("trying %s:%u", p->file, p->line));
		if (match_cond_list(p->cond, qp))
			break;
	}
	return p;
}

void
route_query(const char *id, struct smap_conninfo const *conninfo,
	    smap_stream_t ostr, const char *map, const char *key)
{
	struct smap_database_instance *dbi;
	struct smap_module *mod;
	struct query_pack query;
	struct query_route *qr;
	
	debug(DBG_QUERY, 1, ("routing query %s %s", map, key));
	query.server_id = id;
	query.conninfo = conninfo;
	query.map = map;
	query.key = key;
	qr = find_query_route(&query);
	if (qr)
		debug(DBG_QUERY, 1, ("route at %s:%u, database %s",
				     qr->file, qr->line, qr->dbname));
	else {
		smap_error("no match for %s %s", map, key);
		smap_stream_printf(ostr, "NOTFOUND\n");
		return;
	}

	dbi = qr->dbi;
	mod = dbi->inst->module;
	if (!dbi->opened) {
		int rc = 0;
		
		debug(DBG_DATABASE, 2, ("opening database %s", dbi->id));
		if (mod->smap_open)
			rc = mod->smap_open(dbi->dbh);
		if (rc) {
			smap_error("cannot open database %s", dbi->id);
			/* FIXME: mark database unusable */
			smap_stream_printf(ostr, "NOTFOUND\n");
			return;
		}
		dbi->opened = 1;
	}
	mod->smap_query(dbi->dbh, ostr, map, key, conninfo);
}
